//*************************************************************************
//Declarations and Global variables for Adjunct Tables (programmed by NXH)*
//(version 3.0)                              				  *
//*************************************************************************
#ifndef __ADTABLE__
#define __ADTABLE__
#include<iostream.h>
#include<iomanip.h>
#include<string.h>
#include<assert.h>
#include<stdlib.h>
#include "gconst.h"
#include "random.h"
#include "stree.h"
#include "lexicon.h"
//********************************************************
//Data structure and operations for list of size numbers *
//********************************************************
struct _snode {
	       SIZE id;
	       double prob; //probability
	       struct _snode *next;
	      };
typedef struct _snode snode;
struct _slist {
		snode *head;
		SIZE card;
	       };
typedef struct _slist slist;
//*****************************
//Initialize a list of numbers*
//*****************************
void IniSlist(slist &s)
{
  s.head=NULL;
  s.card=0;
}
//***************************
//Check if the list is emty *
//***************************
BOOL EmptySlist(slist s)
{
 if (s.head==NULL)
  return TRUE;
 else return FALSE;
}
//***********************************
//Return the cardinality of the list*
//***********************************
SIZE CardSlist(slist s)
{
 return s.card;
}
//******************************
//Destroy a list               *
//******************************
void DestroySlist(slist &s)
{
 snode *t, *p;
 t=s.head;
 while (t!=NULL)
 {
  p=t;
  t=t->next;
  delete p;
 }
 s.head=NULL;
 s.card=0;
}
//*****************************
//Add to the head of a list   *
//*****************************
void AddHeadSlist(slist &s,SIZE k,double prob)
{
 snode *t;
 t=new snode; assert(t!=NULL);
 t->id=k;
 t->next=s.head;
 t->prob=prob;
 s.head=t;
 s.card++;
}
//*********************************************
//This function deletes the node p in the list*
//*********************************************
void DeleteNodeSlist(slist &s, snode *p)
{
 snode *t;
 if(p==s.head)
  s.head=(s.head)->next;
 else
  {
   t=s.head;
   while(t->next!=p)
    t=t->next;
   t->next=p->next;
  }
  delete p;
  s.card--;
 if(s.card==0)
  s.head=NULL;
}
//************************
//Normalise Probability  *
//************************
void NormaliseTD(slist s)
{
  double sum;
  snode *t;
  sum=0;
  t=s.head;
  while(t)
   {
    sum+=t->prob;
    t=t->next;
   }
  t=s.head;
  while(t)
   {
    t->prob/=sum;
    t=t->next;
   }
}
//**************************************************
//Get a tree-id from the source list according to  *
//the probability distribution                     *
//**************************************************
SIZE GetTreeID(slist s)
{
 snode *p;
 double h;
 assert(!EmptySlist(s));
 h=Next_Double();
 p=s.head;
 while((h>0)&&p)
  {
   h=h-p->prob;
   if(h>0)
    p=p->next;
  }
 if(p)
  return p->id;
 else
  {
   cout<<endl<<"Adjunction distribution error!!!";
   exit(1);
  }
 return 0;
}
//********************************
//This function display a list   *
//********************************
void DisplaySlist(slist s)
{
 snode *t;
 t=s.head;
 cout<<endl<<"number of trees:"<<s.card<<endl;
 while (t)
 {
   cout<<t->id;
   cout<<"-probability:"<<t->prob<<"; ";
   t=t->next;
 }
}
//******************************************************************
//Data structures and operations for address and address list      *
//******************************************************************
struct _address {
		  slist source;
		  node *addr;
		  node *parent;
		  int  constraint;
		  struct _address *next;
		};
typedef struct _address address;
struct  _addlist{
		  address *head;
		  BYTE card;
		};
typedef struct _addlist addlist;
//***********************************************
//this function initialize the list of address  *
//***********************************************
void IniAddList(addlist &al)
{
 al.head=NULL;
 al.card=0;
}
//***********************************************
//This function check whether the adlist is emty*
//***********************************************
BOOL EmptyAddList(addlist al)
{
 if (al.head==NULL)
  return TRUE;
 else
  return FALSE;
}
//*********************************************
//This function destroy the addlist           *
//*********************************************
void DestroyAddList(addlist &al)
{
 address *t,*p;
 t=al.head;
 while (t!=NULL)
  {
   p=t;
   t=t->next;
   delete p;
  }
  al.head=NULL;
  al.card=0;
}
//*****************************
//Add to the head of a list   *
//*****************************
void AddHeadAddList(addlist &al,slist s,node *addr,node *parent,int constr)
{
 address *t;
 t=new address; assert(t!=NULL);
 t->addr=addr;
 t->parent=parent;
 t->source=s;
 t->constraint=constr;
 t->next=NULL;
 t->next=al.head;
 al.head=t;
 al.card++;
}
void AddTailAddList(addlist &al,slist s,node *addr,node *parent, int constr)
{
address *t,*p;
 t=new address; assert(t!=NULL);
 t->addr=addr;
 t->parent=parent;
 t->source=s;
 t->constraint=constr;
 t->next=NULL;
 if (EmptyAddList(al))
  al.head=t;
 else
  {
   p=al.head;
   while(p->next)
    p=p->next;
   p->next=t;
  }

 al.card++;
}
//*********************************************
//This function deletes the node p in the list*
//*********************************************
void DeleteNodeAddList(addlist &al, address *p)
{
 address *t;
 if(p==al.head)
  al.head=(al.head)->next;
 else
  {
   t=al.head;
   while(t->next!=p)
    t=t->next;
   t->next=p->next;
  }
  delete p;
  al.card--;
 if(al.card==0)
  al.head==NULL;
}
//********************************
//This function display a list   *
//********************************
void DisplayAddList(addlist al)
{
 address *t;
 t=al.head;
 cout<<endl<<"number of address:"<<(int) al.card<<endl;
 while (t)
  {
   DisplaySlist(t->source);
   cout<<endl;
   cout<<"Constraint:"<<t->constraint<<endl;
   cout<<"*************************"<<endl;
   t=t->next;
  }
}
//****************************************************
//declarations and operations for the adjunct table  *
//****************************************************
addlist *adtable;
SIZE  entrynum=0;
//**********************************************
//this function initializes the adjunct table  *
//**********************************************
void InitializeAdTable(void)
{
  SIZE i;
  entrynum=numetree;//each elementary tree has exactly one entry
  adtable= new addlist[entrynum]; assert(adtable!=NULL);
  for(i=0;i<entrynum;i++)
     IniAddList(adtable[i]);
}
//*************************************************************
//Add an address to an adjunction table entry from a tree     *
//*************************************************************
void AddAddress(addlist &al, node *t,node *parent,int constr)
{
 SIZE i;
 node *p;
 slist s;
 if(t->children!=NULL) //exclude the foot node from being adjoined
   {
      IniSlist(s);
     for (i=0;i<numetree;i++)
      if (!IsAlpha(etree[i])&&!strcmp(t->name,(etree[i].root)->name))
//will be changed (inthe next version when adjoined constraints are allowed)
	  AddHeadSlist(s,i,1.0);//can change later
      if (!EmptySlist(s))
	AddHeadAddList(al,s,t,parent,constr);//make the address increasing
   }
}
//*****************************************************
//This function build i-th entry in the adjunct table *
//*****************************************************
void AddEntry(addlist &al,node *t,node *parent)
{
 node *p;
 if (t!=NULL)
 {
   AddAddress(al,t,parent,0);//constraint- 0  no constraint (in this version)
   p=t->children;
   while (p)
   {
     AddEntry(al,p,t);
     p=p->sibling;
   }
 }
}
void NormaliseAD(void)
{
 SIZE i;
 address *a;
 for(i=0;i<entrynum;i++)
 {
  a=adtable[i].head;
  while(a)
  {
   NormaliseTD(a->source);
   a=a->next;
  }
 }
}
//*******************************************
//Build the adjunction table                *
//*******************************************
void BuildAdTable(void)
{
 SIZE i;
  InitializeAdTable();
  for (i=0;i<entrynum;i++)
    AddEntry(adtable[i],etree[i].root,NULL);
  NormaliseAD();
}
#endif
