//***********************************************************************
//This file contains declarations and operations for lexicons in LTAGs  *
//For TAG3P version 3.0. Programmed by NXH			        *
//***********************************************************************
#ifndef __LEXICON__
#define __LEXICON__
#include<iostream.h>
#include<string.h>
#include<iomanip.h>
#include<math.h>
#include<stdlib.h>
#include<assert.h>
#include "random.h"
#include "stree.h"
#define MAXLEX 20
typedef unsigned char LEXSIZE;
struct _lexeme{
			 node *lex;
			 double prob;
	      };
typedef struct _lexeme lexeme;
struct _lexicon  {
		   lexeme lexemes[MAXLEX];
		   LEXSIZE card;
		   char name[MAXNAME];
		};
typedef struct _lexicon lexicon;
lexicon *vocab;
LEXSIZE lexnum;//number of lexicons
struct _lnode {
			LEXSIZE lexiconid;
			LEXSIZE lexemeid;
                        double value;  
		      struct _lnode *next;
		      };
typedef  struct _lnode lnode;
struct _llist  {
		     lnode *head;
		};
typedef struct _llist llist;
void IniLList(llist &l)
{
  l.head=NULL;
}
BOOL EmptyLList(llist l)
{
  if(l.head==NULL)
   return TRUE;
 else
   return FALSE;
}
void DeleteLList(llist  &l)
{
  lnode *p,*q;
  p=l.head;
  while(p)
  {
    q=p;
    p=p->next;
    delete q;
  }
 l.head=NULL;
}
void DeleteHeadLList(llist &l)
{
 lnode *t;
 if(!EmptyLList(l))
  {
   t=l.head;
   l.head=(l.head)->next;
   delete t;
  }
}
void AddTailLList(llist &l,LEXSIZE lexiconid,LEXSIZE lexemeid, double value)
{
   lnode *t,*p;
   t= new lnode; assert(t!=NULL);
   t->lexiconid=lexiconid;
   t->lexemeid=lexemeid;
   t->value=value; 
   t->next=NULL;
 if(EmptyLList(l))
  l.head=t;
 else
   {
     p=l.head;
    while(p->next)
       p=p->next;
     p->next=t;
   }
}
llist CopyLList(llist l)
{
 llist t;
 lnode *p;
 IniLList(t);
 p=l.head;
 while(p)
  {
   AddTailLList(t,p->lexiconid,p->lexemeid,p->value);
   p=p->next;
  }
 return t;
}
//**********************************
//Normalise lexeme distribution    *
//**********************************
void NormaliseLD(LEXSIZE lexiconid)
{
 LEXSIZE i;
 double sum;
 sum=0;
 for(i=0;i<vocab[lexiconid].card;i++)
   sum+=vocab[lexiconid].lexemes[i].prob;
 for(i=0;i<vocab[lexiconid].card;i++)
  vocab[lexiconid].lexemes[i].prob/=sum;
}
void NormaliseVD(void)
{
 LEXIZE i;
 for(i=0;i<lexnum;i++)
 NormaliseLD(i);
}
//*****************************************************************
//Get a lexeme for a lexicon according to the lexeme distribution *
//*****************************************************************
node *GetLexeme(LEXSIZE lexid)
{
  node *s;
  LEXSIZE i;
  double r;
 r=Next_Double();
 i=0;
 while((r>0)&&(i<vocab[lexid].card))
 {
   r-=vocab[lexid].lexemes[i].prob;
   if(r>0)
    i++;
 }
 if(i<vocab[lexid].card)
 {
  _oldfoot=NULL;
  s=CopyNode(vocab[lexid].lexemes[i].lex);
  return s;
 }
 else
  {
   cout<<endl<<"Lexeme Distribution Error!!!";
   exit(1);
  }
  return NULL;
}
LEXIZE GetLexemeID(LEXSIZE lexid)
{
  LEXSIZE i;
  double r;
 r=Next_Double();
 i=0;
 while((r>0)&&(i<vocab[lexid].card))
 {
   r-=vocab[lexid].lexemes[i].prob;
   if(r>0)
    i++;
 }
 if(i<vocab[lexid].card)
  return i;
 else
  {
   cout<<endl<<"Lexeme Distribution Error!!!";
   exit(1);
  }
 return 0;
}
double (*GetERconst)(char *name);
double DefaultConstant(char *name)
{
 return VOIDVALUE;
}
//********************************************************
//Build the list of lexicon for substitition from a tree *
//********************************************************
void BuildLListNode(node *s,llist &l)
{
  LEXIZE i,j;
  node *p;
  double value;
  if(s)
   {
    if(s->children==NULL)//lexicon only at leaves
     {
      // find in the vocab
      i=0;
      while((i<lexnum)&&strcmp(s->name,vocab[i].name))
       i++;
      if(i<lexnum)//found
      {
       j=GetLexemeID(i);
       if((vocab[i].lexemes[j].lex)->name[0]=='R') //ephemeral constant  
        value=GetERconst((vocab[i].lexemes[j].lex)->name);
       else
        value=VOIDVALUE;   
       AddTailLList(l,i,j,value);
      }
     }
     else
     {
      p=s->children;
      while(p)
       {
	BuildLListNode(p,l);
	p=p->sibling;
       }
     }
   }
}
llist BuildLList(node *t)
{
 llist l;
 IniLList(l);
 BuildLListNode(t,l);
 return l;
}
//*************************
//Initialisation of vocab;*
//*************************

//*********************************************************************
//Substitute a s-tree with lexicons according to lexeme distributions *
//*********************************************************************
void SubstituteLexicon(node *s)
{
  node *p,*lex;
  LEXSIZE i;
if(s)
{
  //the lexicon must be at leaves
 if(s->children==NULL)
  {
    //check if it is in the vocab
     i=0;
    while((i<lexnum) &&strcmp(s->name,vocab[i].name))
       i++;
     if(i<lexnum) //if found
     {
	lex=GetLexeme(i);
	strcpy(s->name,lex->name);
      if(s->name[0]=='R')
       s->value=GetERconst(lex->name);
      else
       s->value=VOIDVALUE;
	s->children=lex->children;
	delete lex;

     }

  }
 else
  {
     p=s->children;
    while(p)
       {
	  SubstituteLexicon(p);
	  p=p->sibling;
       }
   }

 }
}
//**********************************************
//Substitute a s-tree with lexicon from a list *
//**********************************************
void DSubstituteLexiconNode(node *s, llist &l)
{
 node *p, *lex;
 lnode *t;
 BOOL found;
 if(s)
  {
    if(s->children==NULL) //lexicon only at leaves
     {
       //find the lexicon at the head
       t=l.head;
	if(t)
	 if(!strcmp(s->name,vocab[t->lexiconid].name))
	  {
	     _oldfoot=NULL;
	     lex=CopyNode(vocab[t->lexiconid].lexemes[t->lexemeid].lex);
	     strcpy(s->name,lex->name);
	     if(s->name[0]=='R')
               s->value=t->value;
             s->children=lex->children;
	     delete lex;
	     DeleteHeadLList(l);
	  }
     }
   else
     {
	p=s->children;
       while(p)
	 {
	   DSubstituteLexiconNode(p,l);
	   p=p->sibling;
	 }
     }
  }
}
void DSubstituteLexicon(node *s, llist l)
{
 llist temp;
 temp=CopyLList(l);
 DSubstituteLexiconNode(s,temp);
 DeleteLList(temp);
}
//*****************************
//Initialise the vocabulary   *
//*****************************
#endif
