#ifndef __STREE__
#define __STREE__
//**********************************************************************
//This file contains declaration and utility to manipulate sentencial                                      *
//trees. Programmed by NXH. (version 3.0)                                                                        *
//**********************************************************************
#include<iostream.h>
#include<string.h>
#include<assert.h>
#include<ctype.h>
#include<stdlib.h>
#include "gconst.h"

#define VOIDVALUE -999999999.2732897
//*********************************************************************
//Declarations and Functions for Node trees                           *
//*********************************************************************
struct _node{
	      char name[MAXNAME];
            double value;
	      struct _node *sibling, *children;
	   };
typedef struct _node node;
//***************************
//Make a Node               *
//***************************
node *MakeNode(char *nname,double value)
{
  node *t;
  t=new node;
  assert(t!=NULL);
  strcpy(t->name,nname);
  t->value=value;
  t->sibling=NULL;
  t->children=NULL;
  return t;
}
//*****************************
//Delete a subtree            *
//*****************************
void DeleteNode(node *t)
{
  node *q,*p;
  if (t)
   {
    if(t->children==NULL)
      delete t;
     else
      {
	p=t->children;
	while(p)
	  {
	    q=p->sibling;
	    DeleteNode(p);
	    p=q;
	  }
	 delete t;
      }
  }
}
//**********************************
//Copy a subtree                   *
//**********************************
node *_oldfoot=NULL;//this variable contains the foot node of the source tree
node *_newfoot=NULL;//this variable contains the footnode of the newtree
node *CopyNode(node *s)
{
  node *t,*p,*q;
  if(s)
   {
     t=MakeNode(s->name,s->value);
     if (s==_oldfoot)
	 _newfoot=t;
     q=s->children;
     t->children=CopyNode(q);
     p=t->children;
     while (q)
      {
	p->sibling=CopyNode(q->sibling);
	q=q->sibling;
	p=p->sibling;
      }
     return t;
   }
  return NULL;
}
//***********************************
//Get the height of a subtree       *
//***********************************
int GetHeightNode(node *t)
{
   if (t->children==NULL)
     return 0;
   else
     {
       unsigned int c,max=0;
       node *p=t->children;
       while (p)
	{
	  c=GetHeightNode(p);
	  if (c>max)
	      max=c;
	  p=p->sibling;
	 }
	return max+1;
      }
}
//*************************************
//Compare two subtree                 *
//*************************************
BOOL NodeCmp(node *t,node *s)
{
 if((t!=NULL) && (s!=NULL))
  {
  if(!strcmp(t->name,s->name)&&(t->value==s->value))
   {
     node *p=t->children,*q=s->children;
     BOOL b=TRUE;
     while(b && (p!=NULL) && (q!=NULL))
      {
	b=NodeCmp(p,q);
	p=p->sibling;
	q=q->sibling;
      }
      if (b==FALSE)
	  return b;
       else
	  if ((p==NULL)&&(q==NULL))
	   return TRUE;
	  else
	   return FALSE;
   }
   else
    return FALSE;
  }
 else
  if((t==NULL) && (s==NULL))
   return TRUE;
    else
     return FALSE;
}
//*********************************************************
//Travel a Tree in Preorder,Inorder and Postoder          *
//*********************************************************
void (*_doit)(node *n);
void Preorder(node *t)
{
   node *p;
   if(t)
    {
     _doit(t);
      p=t->children;
      while (p)
       {
	 Preorder(p);
	 p=p->sibling;
       }
    }
}
//Travel a Tree in Inoder
void Inorder(node *t)
{
  node *p;
  if (t)
   {
     p=t->children;
     Inorder(p);
     _doit(t);
     while (p)
     {
      p=p->sibling;
      Inorder(p);
     }
   }
}
//Travel a Tree in Postorder
void Postorder(node *t)
{
 node *p;
 if (t)
  {
    p=t->children;
    while (p)
     {
      Postorder(p);
      p=p->sibling;
     }
    _doit(t);
  }
}
//***********************************
//From a string to STree            *
//***********************************
char _sbuff[MAXSTRING];
int  _curchar=0;
node *_sfoot=NULL;
node *StringToTree()
{
   char n[MAXNAME];
   int i;
   node *t;
   while(_sbuff[_curchar]==' ')
     _curchar++;
    if(isalnum(_sbuff[_curchar]))
     {
	i=0;
	while (isalnum(_sbuff[_curchar])&&(i<MAXNAME-1))
	 {
	   n[i]=_sbuff[_curchar];
	   i++;_curchar++;
	 }
	n[i]=0;//terminate the string
	t=MakeNode(n,VOIDVALUE);
	if (_sbuff[_curchar]=='~')
	 {
	   _sfoot=t;
	   _curchar++;
	 }
       while(_sbuff[_curchar]==' ')
	_curchar++;
       switch (_sbuff[_curchar]){
	case '(':
		_curchar++;
		t->children=StringToTree();
		_curchar++;
		while(_sbuff[_curchar]==' ')
		 _curchar++;
	       //finding for the next sibling
		if(_sbuff[_curchar]==',')
		  {
		    _curchar++;
		    t->sibling=StringToTree();
		  }
		return t;
	case ')':
	       //_curchar++;
		return t;
	case ',':
		_curchar++;
		t->sibling=StringToTree();
		return t;
	case 0:
		return t;//new part
	default:
	      cout<<"Error in the string";
	      exit(1);
      }
     }
     else
     {
       cout<<"Error in the string";
       exit(1);
     }
    return NULL;
 }
//*************************************
//From a tree to string with brackets *
//*************************************
void TreeToString(node *r,node *f)
{
 node *p;
 char st[20]; 
 if (r)
   {
    if(r->name[0]=='R')
     {
      strcpy(st,"");
      sprintf(st,"%lf",r->value);
      strcat(_sbuff,st);
     }
    else
     strcat(_sbuff,r->name);
    if (r==f)
     strcat(_sbuff,"~");
    if(r->children)
     {
       strcat(_sbuff,"(");
       p=r->children;
       while(p)
	{
	  TreeToString(p,f);
	  p=p->sibling;
	  if (p)
	   strcat (_sbuff,",");
	}
       strcat(_sbuff,")");
     }
   }
}
//*****************************************************
//To check if node p is in the subtree stermed from q *
//*****************************************************
BOOL InNode(node *p, node *q)
{
 node *t;
 t=q;
 if (t==p)
  return TRUE;
 else
  if (t)
   {
    t=t->children;
    while (t)
     if (InNode(p,t))
      return TRUE;
     else
      t=t->sibling;
    }
  return FALSE;
}
//*******************************************************
//This function compare two node it returns             *
//+1 if p is in q,-1 if q is in p, and 0 if incomparable*
//*******************************************************
int NodeOrder(node *p, node *q)
{
 if (InNode(p,q))
  return 1;
 else
  if (InNode(q,p))
    return -1;
  else
   return 0;
}
//********************************
//Compute the size of a subtree  *
//********************************
int  _nodecount;
BOOL _cntleave;
void NodeCount(node *t)
{
 _nodecount++;
 if(!_cntleave && (t->children==NULL))//if don't count the leaves
 _nodecount--;
}
int  GetSizeNode(node *t,BOOL countleave)
{
 _cntleave=countleave;
 _nodecount=0;
 _doit=NodeCount;
 Preorder(t);
 return _nodecount;
}
//************************************************************************
//Declarations and Global Variables for Strees                           *
//************************************************************************

struct  _stree {
		    node *root;
		    node *foot;
	       };
typedef struct _stree stree;
stree *etree;
SIZE  numetree;
SIZE numalpha,numbeta;
//*********************************
//Initialize a Null tree          *
//*********************************
stree  InitEmptyTree()
{
  stree s;
  s.root=NULL;
  s.foot=NULL;
  return s;
}
stree InitTree(const char *st)
{
 stree s;
 strcpy(_sbuff,st);
 _sfoot=NULL;
 _curchar=0;
 s.root=StringToTree();
 s.foot=_sfoot;
 return s;
}
//*********************************
//Delete a tree from memory       *
//*********************************
void DeleteTree(stree s)
{
 DeleteNode(s.root);
 s.root=s.foot=NULL;
}
void Display(stree s)
{
  strcpy(_sbuff,"");
  TreeToString(s.root,s.foot);
  cout<<endl<<_sbuff;
}
//*********************************
//Check whether a tree is empty   *
//*********************************
BOOL IsEmpty(stree s)
{
 return (s.root==NULL);
}
//*****************************************
//Check whether a tree is an alpha tree   *
//*****************************************
BOOL IsAlpha(stree s)
{
  return (s.foot==NULL);
}
//***************************
//Get height of a tree      *
//***************************
int  GetHeight(stree s)
{
  return GetHeightNode(s.root);
}
int GetSize(stree s,BOOL leave)
{
 return GetSizeNode(s.root,leave);
}
//****************************
//Copy a tree                *
//****************************
stree CopyTree(stree s)
{
   stree t;
  _oldfoot=s.foot;
  _newfoot=NULL;
  t.root=CopyNode(s.root);
  t.foot=_newfoot;
  return t;
}
//****************************
//Compare two trees          *
//****************************
BOOL CmpTree(stree t, stree s)
{
  return NodeCmp(t.root,s.root);
}
//**************************************************************
//Adjoin a list of beta trees to an alpha or betatree          *
//**************************************************************
stree AdjointTree(stree beta,stree gama,node *add,node *parent)
{
  stree t,v;
  t=beta;
  v=gama;
  if(parent==NULL)//adjoin at root
    {
      (t.foot)->children=(v.root)->children;
      t.foot=v.foot;
      delete v.root;
      return t;
    }
  else
     {
       node *p,*q;
       p=parent->children;
       q=p;
       while ((p!=add)&& p)
       {
	 q=p;
	 p=p->sibling;
       }
       if (p==q)//adjoin at the first children
	{
	  parent->children=t.root;
	  (t.root)->sibling=p->sibling;
	  (t.foot)->children=p->children;
	  delete p;
	  return v;
	}
       else
	{
	 q->sibling=t.root;
	 (t.root)->sibling=p->sibling;
	 (t.foot)->children=p->children;
	 delete p;
	 return v;
	}
     }
}
#endif
