//************************************************************************
//This module contains fitness case and functions to compute raw fitness *
//as well as standardized fitness for Symbolic Regression Problem        *
//Programmed by NXH.                                                     *
//************************************************************************
#ifndef __SYMBOLICREG__
#define __SYMBOLICREG__

#include<iostream.h>
#include<string.h>
#include<stdio.h>
#include<assert.h>
#include<math.h>
#include<stdlib.h>
#include "gconst.h"
#include "random.h"
#include "stree.h"
#include "lexicon.h"
#include "adtable.h"
#include "dtree.h"

#define NUMFITCASE 20
#define INFPLUS  exp(700)
#define INFMINUS -exp(700)
//#define EXPOVL 650
#define PVAL(y) ((y==HUGE_VAL)?(INFPLUS):((y==-HUGE_VAL)?(-INFMINUS):(y)))
//****************************************
//Fitness cases for symbolic regression  *
//They are sampled between 0  and 1      *
//Seed for sampling 12345                *
//****************************************
struct _sample{
		 double x;
		 double y;
		};
typedef struct _sample sample;
sample fitcase[NUMFITCASE];
BOOL SuccPredicate=FALSE;
void SetFitCase()
{
  double x;
  int i;
  for(i=0;i<NUMFITCASE;i++)
   {
    // fscanf(fp,"%lf%lf",&x,&y);
     fitcase[i].x=2*Next_Double()-1;
     x=fitcase[i].x;
     fitcase[i].y=x*x*x*x+x*x*x+x*x+x;
   }
}
//This function is for getting ephemeral constants
double ERGenerate(char *st)
{
  if(!strcmp(st,"R1"))
   return Next_Double();
  else
  if(!strcmp(st,"R2"))
   return 2*Next_Double()-1;
 else 
 return VOIDVALUE; 
}
//********************************************************************
//This function compute the function (in tree form) with given input *
//********************************************************************
double Compute(node *st,double x)
{
 node *p;
 double l,r;
 if (!strcmp(st->name,"X"))
     return x;
 else //st->name="EXP"
  if(!strcmp(st->name,"R1")||!strcmp(st->name,"R2"))
   return st->value;
  else
  {
    p=st->children;
    if (!strcmp(p->name,"EXP"))
    {
     l=Compute(p,x);//compute the left branch
     p=p->sibling;
     r=Compute(p->sibling,x);//compute the right branch
     p=p->children;//get the operator
     if (!strcmp(p->name,"add"))
       return PVAL(r+l);
     else
      if (!strcmp(p->name,"sub"))
	return PVAL(l-r);
	else
	 if (!strcmp(p->name,"mul"))
	  return PVAL(r*l);
	  else
	    if (r!=0)   //protected division
	     return PVAL(l/r);
	     else
	      return 1;
    }
   else
    if (!strcmp(p->name,"PRE"))
     {
       r=Compute(p->sibling,x); //compute exp inside function
       p=p->children;
       if (!strcmp(p->name,"sin"))
	return sin(r);
       else
	if(!strcmp(p->name,"cos"))
	 return cos(r);
	  else
	  if(!strcmp(p->name,"ep"))
	      return PVAL(exp(r));
	   else
	     if (r!=0)   //protected logarithm
	      return log(fabs(r));
	       else
		return 0;
     }
    else //VAR
      return Compute(p->children,x);
  }
}
//*********************************************
//Compute the raw fitness for an invidual     *
//*********************************************
double ComputeRF(stree st)
{
  double sum=0,t;
  int i;
  int hit;
  hit=0;
  for(i=0;i<NUMFITCASE;i++)
   {
     t=PVAL(fabs(Compute(st.root,fitcase[i].x)-fitcase[i].y));
     if(t<=0.01)
       hit++;
     sum+=t;
   }
   if(hit==NUMFITCASE)
     SuccPredicate=TRUE;
   return sum;
}
void BuildVocab()
{
 char *lx[]={"R1","R2"};
 stree s;
 vocab=new lexicon[1];
 lexnum=1; 
 strcpy(vocab[0].name,"R");
 vocab[0].card=2;
 s=InitTree(lx[0]);
 vocab[0].lexemes[0].lex=s.root;
 vocab[0].lexemes[0].prob=1.0; 
 s=InitTree(lx[1]);
 vocab[0].lexemes[1].lex=s.root;
 vocab[0].lexemes[1].prob=1.0; 
 NormaliseVD();

 
}
//****************************************************************
//This function initializes elementary trees and adjoining table *
//****************************************************************
void LoadEtree()
{
 cout<<endl<<"Loading elementary trees...";
 etree=new stree[20];assert(etree!=NULL);
 etree[0]=InitTree("EXP(VAR(X))");
 etree[1]=InitTree("EXP(VAR(R))");
 etree[2]=InitTree("EXP(EXP(VAR(X)),OP(add),EXP~)");
 etree[3]=InitTree("EXP(EXP(VAR(X)),OP(mul),EXP~)");
 etree[4]=InitTree("EXP(EXP(VAR(X)),OP(sub),EXP~)");
 etree[5]=InitTree("EXP(EXP(VAR(X)),OP(div),EXP~)");
 etree[6]=InitTree("EXP(EXP~,OP(add),EXP(VAR(X)))");
 etree[7]=InitTree("EXP(EXP~,OP(mul),EXP(VAR(X)))");
 etree[8]=InitTree("EXP(EXP~,OP(sub),EXP(VAR(X)))");
 etree[9]=InitTree("EXP(EXP~,OP(div),EXP(VAR(X)))");
 etree[10]=InitTree("EXP(EXP(VAR(R)),OP(add),EXP~)");
 etree[11]=InitTree("EXP(EXP(VAR(R)),OP(mul),EXP~)");
 etree[12]=InitTree("EXP(EXP(VAR(R)),OP(sub),EXP~)");
 etree[13]=InitTree("EXP(EXP(VAR(R)),OP(div),EXP~)");
 etree[14]=InitTree("EXP(EXP~,OP(add),EXP(VAR(R)))");
 etree[15]=InitTree("EXP(EXP~,OP(mul),EXP(VAR(R)))");
 etree[16]=InitTree("EXP(EXP~,OP(sub),EXP(VAR(R)))");
 etree[17]=InitTree("EXP(EXP~,OP(div),EXP(VAR(R)))");

 numetree=18;
 numalpha=2;
 numbeta=16;
 cout<<"Done!";
 cout<<endl<<"Building Vocabulary...";
 BuildVocab();
 cout<<"done";
 cout<<endl<<"Buiding adjoining table...";
 entrynum=0;
 BuildAdTable();
 cout<<"Done!";
}
node *Stree2ParseTree(node *st)
{
 node *p,*q,*t,*t1,*t2;
 if(!strcmp(st->name,"EXP"))
  {
   p=st->children;
   if(!strcmp(p->name,"EXP"))//binary oprators
    {
     //get the oprators
     q=p->sibling->children;
     t=MakeNode(q->name,q->value);
     t1=Stree2ParseTree(p);
     t2=Stree2ParseTree(p->sibling->sibling);
     t->children=t1;
     t1->sibling=t2;
     t2->sibling=NULL;
     return t;
    }
    else
     if(!strcmp(p->name,"PRE"))
     {
      q=p->children;
      t=MakeNode(q->name,q->value);
      t1=Stree2ParseTree(p->sibling);
      t->children=t1;
      t1->sibling=NULL;
      return t;
     }
    else //VAR
     return Stree2ParseTree(p->children);
  }
 else //"X"
  {
    t=MakeNode(st->name,st->value);
    return t;
  }

}
#endif
